# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""A module that contains some forecasting models: ARIMA, Prophet."""

from abc import abstractmethod
from typing import List, Tuple, Any, Optional, Dict, Union, cast

import logging
import numpy as np
import os
import pandas as pd
import uuid

from joblib import Parallel, delayed

from azureml.automl.core.shared import (
    forecasting_utils,
    time_series_data_frame,
    constants,
    model_wrappers,
    exceptions,
    import_utilities,
    utilities)
from azureml.automl.core.shared.types import GrainType
from azureml.automl.core.shared.exceptions import FitException


class _MultiGrainForecastBase:
    """
    Multi-grain forecast base class.

    Enables multi-grain fit and predict on learners that normally can only operate on a single timeseries.
    """

    def __init__(self,
                 timeseries_param_dict: Dict[str, Any]):
        self.timeseries_param_dict = timeseries_param_dict
        self.time_column_name = self.timeseries_param_dict[constants.TimeSeries.TIME_COLUMN_NAME]
        self.grain_column_names = self.timeseries_param_dict.get(constants.TimeSeries.GRAIN_COLUMN_NAMES, [])
        self.grain_column_names = [constants.TimeSeriesInternal.DUMMY_GRAIN_COLUMN] if \
            self.grain_column_names is None or len(self.grain_column_names) == 0 else self.grain_column_names
        self.drop_column_names = self.timeseries_param_dict.get(constants.TimeSeries.DROP_COLUMN_NAMES, []) or []
        self._max_cores_per_iteration = self.timeseries_param_dict.get(constants.TimeSeries.MAX_CORES_PER_ITERATION, 1)

        # model state
        self._grain_levels = []  # type: List[Tuple[str]]
        self._models = {}  # type: Dict[Tuple[str], Any]
        self._last_observation_dates = {}  # type: Dict[Tuple[str], pd.Timestamp]
        self._first_observation_dates = {}  # type: Dict[Tuple[str], pd.Timestamp]
        self._freq = None  # type: Optional[pd.DateOffset]
        self._is_fit = False

        # what to predict
        self._quantiles = [.5]

    # TODO: duplicates code from RegressionPipeline
    # /src/automl_client_core_runtime/automl/client/core/runtime/model_wrappers.py
    # Perhaps we should make a QuantileMixin.
    @property
    def quantiles(self) -> List[float]:
        """Quantiles for the model to predict."""
        return self._quantiles

    @quantiles.setter
    def quantiles(self, quantiles: Union[float, List[float]]) -> None:
        if not isinstance(quantiles, list):
            quantiles = [quantiles]

        for quant in quantiles:
            if quant == 0:
                raise FitException(
                    "Quantile 0 is not supported.", target="quantiles",
                    reference_code="forecasting_models._MultiGrainForecastBase.quantiles.equal_0",
                    has_pii=False)
            if quant == 1:
                raise FitException(
                    "Quantile 1 is not supported.", target="quantiles",
                    reference_code="forecasting_models._MultiGrainForecastBase.quantiles.equal_1",
                    has_pii=False)
            if quant < 0 or quant > 1:
                raise FitException(
                    "Quantiles must be strictly less than 1 and greater than 0.", target="quantiles",
                    reference_code="forecasting_models._MultiGrainForecastBase.quantiles.out_of_range",
                    has_pii=False)

        self._quantiles = quantiles

    @staticmethod
    def _get_num_parallel_process(
            len_grain_levels: int,
            max_cores_per_iteration: Optional[int]) -> int:
        """Get num of process for joblib parallelism when fitting models for the grains.

        :param len_grain_levels: length of grain_levels list
        :type len_grain_levels: int
        :param max_cores_per_iteration: max_cores_per_iteration parameter input from timeseries_param_dict
        :type max_cores_per_iteration: Optional[int]
        :return: number of process used for joblib parallelism
        :rtype: int
        """
        cpu_cnt = os.cpu_count()
        num_par_process = 1
        if cpu_cnt is not None:
            num_par_process = max(1, cpu_cnt)
        if max_cores_per_iteration is not None and max_cores_per_iteration != -1:
            num_par_process = min(num_par_process, max_cores_per_iteration)
        num_par_process = min(num_par_process, len_grain_levels)

        return num_par_process

    def fit(self,
            X: pd.DataFrame,
            y: np.ndarray,
            **kwargs: Any) -> None:
        """Fit the model.

        :param X: Training data.
        :type X: pd.DataFrame
        :param y: Training label
        :type y: np.ndarray
        :return: Nothing
        :rtype: None
        """
        tsdf = self._construct_tsdf(X, y)

        tsdf_bygrain = tsdf.groupby_grain()
        self._grain_levels = list(tsdf_bygrain.groups)

        # Initialize the models and state variables
        self._models = {lvl: None for lvl in self._grain_levels}
        self._last_observation_dates = {
            lvl: None for lvl in self._grain_levels}
        self._first_observation_dates = {
            lvl: None for lvl in self._grain_levels}

        # compute number of parallel process in this scenario
        num_par_process = self._get_num_parallel_process(
            len(self._grain_levels), self._max_cores_per_iteration)

        # if num_par_process ==1, bypass the joblib parallel fitting code since it
        # just introduces overhead
        if num_par_process == 1:
            for lvl, series_frame in tsdf_bygrain:
                lvl, first, last, model = _MultiGrainForecastBase._fit_single_grain(self, lvl, series_frame)
                self._first_observation_dates[lvl] = first
                self._last_observation_dates[lvl] = last
                self._models[lvl] = model

        # if num_par_process >1, parallel model fitting for each grain
        # Note, we need to copy the data frame, obtained from the groupby
        # object, because it is implicitly generating a weakref object
        # and it will cause the pikling error when joblib will create the
        # new process.
        if num_par_process > 1:
            results = Parallel(n_jobs=num_par_process)(delayed(_MultiGrainForecastBase._fit_single_grain)(
                self,
                lvl,
                series_frame.copy()) for lvl, series_frame in tsdf_bygrain)
            # Parse results received from Parallell
            for result in results:
                lvl, first, last, model = result
                self._first_observation_dates[lvl] = first
                self._last_observation_dates[lvl] = last
                self._models[lvl] = model

        self._freq = tsdf.infer_freq()
        self._is_fit = True

    def predict(self, X: pd.DataFrame) -> np.ndarray:
        if not self._is_fit:
            raise exceptions.UntrainedModelException()

        tsdf = self._construct_tsdf(X)

        max_horizons = self._get_forecast_horizons(tsdf)
        # Make a dataframe of forecasts
        fcast_df = self._get_forecast(tsdf, max_horizons)

        # Get rows containing in-sample data if any
        in_sample_data = pd.DataFrame()
        in_sample_dfs = []  # type: List[pd.DataFrame]
        for g, X_group in tsdf.groupby_grain():
            if g in self._grain_levels:
                in_sample_dfs.append(X_group.loc[X_group.time_index <= self._last_observation_dates[g]])
        in_sample_data = pd.concat(in_sample_dfs)
        del in_sample_dfs

        # Get fitted results for in-sample data
        if in_sample_data.shape[0] > 0:
            in_sample_fitted = self._fit_in_sample(in_sample_data)
            in_sample_fitted = in_sample_fitted.loc[:, fcast_df.columns]
            fcast_df = pd.concat([in_sample_fitted, fcast_df])

        # We're going to join the forecasts to the input - but first:
        # Convert X to a plain data frame and drop the prediction
        #  columns if they already exist
        point_name = constants.TimeSeriesInternal.DUMMY_PREDICT_COLUMN
        X_df = pd.DataFrame(tsdf, copy=False).drop(axis=1,
                                                   labels=[point_name],
                                                   errors='ignore')

        # Left join the forecasts into the input;
        #  the join is on the index levels
        pred_df = X_df.merge(fcast_df, how='left',
                             left_index=True, right_index=True)

        return cast(np.ndarray, pred_df[constants.TimeSeriesInternal.DUMMY_PREDICT_COLUMN].values)

    def _construct_tsdf(
        self,
        X: pd.DataFrame,
        y: Optional[np.ndarray] = None
    ) -> time_series_data_frame.TimeSeriesDataFrame:
        X = X.copy()
        # Add the Dummy grain coumn only if it was not already added.
        if self.grain_column_names == [constants.TimeSeriesInternal.DUMMY_GRAIN_COLUMN] and\
           self.grain_column_names[0] not in X.index.names and self.grain_column_names[0] not in X.columns:
            X[constants.TimeSeriesInternal.DUMMY_GRAIN_COLUMN] = constants.TimeSeriesInternal.DUMMY_GRAIN_COLUMN

        tsdf_kwargs = {'grain_colnames': self.grain_column_names}
        if y is not None:
            X[constants.TimeSeriesInternal.DUMMY_TARGET_COLUMN] = y
            tsdf_kwargs['ts_value_colname'] = constants.TimeSeriesInternal.DUMMY_TARGET_COLUMN

        # Handle the situation, when we have the time index with the same name as column.
        temp_column_name = None
        if isinstance(X.index, pd.DatetimeIndex) and X.index.names[0] == self.time_column_name:
            # DatetimeIndex is one column representing datetime.
            # We will temporary rename it and after TSDF creation we will revert the change.
            temp_column_name = str(uuid.uuid1())
            X.rename({self.time_column_name: temp_column_name}, inplace=True, axis=1)

        tsdf = time_series_data_frame.TimeSeriesDataFrame(X,
                                                          self.time_column_name,
                                                          **tsdf_kwargs)
        if temp_column_name is not None:
            tsdf.rename({temp_column_name: self.time_column_name}, inplace=True, axis=1)

        return tsdf

    def _fit_in_sample(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Return the fitted values from a the RecursiveForecaster model.

        :param X:
            A TimeSeriesDataFrame defining the data for which fitted values
            are desired.  Inputting the same data used to fit the model will
            return all fitted data.
        :type X: TimeSeriesDataFrame

        :Returns:
            a ForecastDataFrame containing the fitted values in `pred_point`.
        """
        point_name = constants.TimeSeriesInternal.DUMMY_PREDICT_COLUMN
        origin_name = constants.TimeSeriesInternal.ORIGIN_TIME_COLUMN_NAME

        fitted_df = pd.DataFrame()
        for g, X_grain in X.groupby_grain():
            origin_time = self._last_observation_dates[g]
            fitted = self._fit_in_sample_single_grain_impl(self._models[g], g, X_grain)
            assign_dict = {origin_name: origin_time,
                           point_name: fitted}
            X_grain = X_grain.assign(**assign_dict)

            fitted_df = pd.concat([fitted_df, X_grain])

        fitted_df = fitted_df.loc[X.index, :]

        return fitted_df

    def _get_forecast_horizons(self, X: time_series_data_frame.TimeSeriesDataFrame) -> Dict[Tuple[Any], int]:
        """
        Find maximum horizons to forecast in the prediction frame X.

        Returns a dictionary, grain -> max horizon.
        Horizons are calculated relative to the latest training
        dates for each grain in X.
        If X has a grain that isn't present in the training data,
        this method returns a zero for that grain.
        """
        # Internal function for getting horizon for a single grain

        import warnings

        def horizon_by_grain(gr, Xgr):
            try:
                horizon = len(pd.date_range(start=self._last_observation_dates[gr],
                                            end=Xgr.time_index.max(),
                                            freq=self._freq)) - 1
                # -1 because this will INCLUDE the
                # last obs date
            except KeyError:
                horizon = 0

            return horizon
        # ------------------------------------------

        fcast_horizon = {gr: horizon_by_grain(gr, Xgr)
                         for gr, Xgr in X.groupby_grain()}

        negatives = [h <= 0 for h in list(fcast_horizon.values())]
        if any(negatives):
            warnings.warn(('Non-positive forecast horizons detected. Check data for time '
                           'overlap between train and test and/or grains in test data '
                           'that are not present in training data. Failures may occur.'))

        return fcast_horizon

    def _get_forecast(self,
                      X: time_series_data_frame.TimeSeriesDataFrame,
                      max_horizon: Dict[Tuple[Any], int]) -> pd.DataFrame:
        """
        Generate forecasts up to max_horizon for each grain in X.

        The max_horizon parameter can be a single integer or
        a dictionary mapping each grain in X to an integer.

        Returns a pandas DataFrame. The index of this data frame
        will have the same levels as the input, X.
        The ouput will have the following:
        time, grain(s), origin time, point forecast.
        """
        # Get column names from X
        point_name = constants.TimeSeriesInternal.DUMMY_PREDICT_COLUMN
        origin_time_colname = constants.TimeSeriesInternal.ORIGIN_TIME_COLUMN_NAME

        grouped = X.groupby_grain()

        # Make max_horizon forecasts for each grain
        # Note: the whole prediction dataframe needs to be passed,
        # not just the grain name.

        fcast_df = pd.concat([
            self._get_forecast_single_grain(gr,
                                            grain_ctx.loc[grain_ctx.time_index > self._last_observation_dates[gr]],
                                            max_horizon[gr],
                                            X.time_colname,
                                            X.grain_colnames,
                                            origin_time_colname,
                                            point_name)
            for gr, grain_ctx in grouped])

        return fcast_df.set_index(X.index.names)

    def _get_forecast_single_grain(self,
                                   grain_level: Tuple[str],
                                   grain_ctx: pd.DataFrame,
                                   max_horizon: int,
                                   time_colname: str,
                                   grain_colnames: List[str],
                                   origin_time_colname: str,
                                   pred_point_colname: str) -> pd.DataFrame:
        """
        Generate forecasts up to max_horizon for a single grain.

        Returns a plain pandas Dataframe with the following columns:
        time, grain(s), origin time, point forecast,
        distribution forecast (optional).
        """
        if grain_level not in self._grain_levels or not self._models[grain_level]:

            raise exceptions.DataException(model_wrappers.ForecastingPipelineWrapper.FATAL_NO_GRAIN_IN_TRAIN,
                                           has_pii=False)
        # ---------------------------------------------------------------

        # Origin date/time is the latest training date, by definition

        # Note: this does not support the newer forecast interface which
        # allows using a training model away from training data as long
        # as sufficient context is provided.  origin data should instead
        # be computed from the prediction context dataframe (X).
        origin_date = self._last_observation_dates[grain_level]

        # Retrieve the trained model and make a point forecast
        if max_horizon <= 0:
            fcast_dict = {time_colname: np.empty(0),
                          origin_time_colname: np.empty(0),
                          pred_point_colname: np.empty(0)}
        else:
            trained_model = self._models[grain_level]
            point_fcast = self._get_forecast_single_grain_impl(trained_model,
                                                               max_horizon,
                                                               grain_level,
                                                               grain_ctx)

            # Construct the time axis that aligns with the forecasts
            fcast_dates = grain_ctx.index.get_level_values(self.time_column_name)
            fcast_dict = {time_colname: fcast_dates,
                          origin_time_colname: origin_date,
                          pred_point_colname: point_fcast}

        if grain_colnames is not None:
            fcast_dict.update(forecasting_utils.grain_level_to_dict(grain_colnames,
                                                                    grain_level))
        return pd.DataFrame(fcast_dict)

    @staticmethod
    def _fit_single_grain(untrained_model: '_MultiGrainForecastBase',
                          lvl: GrainType,
                          series_frame: time_series_data_frame.TimeSeriesDataFrame,
                          ) -> Tuple[GrainType, pd.Timestamp, pd.Timestamp, Any]:
        """
        Fit the model for a single grain.

        **Note:** this method calls _fit_single_grain_impl internally.
        :param lvl: The grain level.
        :param series_frame: The data frame representing this grain.
        :return: The tuple with grain level, first date, last date and the trained model.
        """
        series_frame.sort_index()
        model = untrained_model._fit_single_grain_impl(series_frame, lvl)

        # Gather the last observation date if time_colname is set
        last = series_frame.time_index.max()
        first = series_frame.time_index.min()
        return lvl, first, last, model

    @abstractmethod
    def _fit_in_sample_single_grain_impl(self,
                                         model: Any,
                                         grain_level: Tuple[Any],
                                         X_grain: time_series_data_frame.TimeSeriesDataFrame) -> np.ndarray:
        """
        Return the fitted in-sample values from a model.

        :param model:
            is an object representation of a model. It is the
            object returned by the _fit_single_grain_impl method.

        :param grain_level:
            is an object that identifies the series by its
            grain group in a TimeSeriesDataFrame. In practice, it is an element
            of X.groupby_grain().groups.keys(). Implementers can use
            the grain_level to store time series specific state needed for
            training or forecasting. See ets.py for examples.
        :param X_grain:
            the context data for the in-sample prediction.

        :param start:
            starting frame of the in sample prediction.

        :param end:
            end frame of the in sample prediction.

        :Returns:
            a 1-D numpy array of fitted values for the training data. The data are
            assumed to be in chronological order
        """
        raise NotImplementedError()

    @abstractmethod
    def _get_forecast_single_grain_impl(self,
                                        model: Any,
                                        max_horizon: int,
                                        grain_level: Tuple[Any],
                                        X_pred_grain: time_series_data_frame.TimeSeriesDataFrame) -> np.ndarray:
        """
        Return the forecasted value for a single grain.

        :param model:
            trained model.
        :param max_horizon:
            int that represents the max horizon.
        :param grain_level:
            tuple that identifies the timeseries the model belongs to.
        :param X_pred_grain
            a dataframe containing the prediction context
        :Returns:
            a 1-D numpy array of fitted values for the training data. The data are
            assumed to be in chronological order
        """
        raise NotImplementedError

    @abstractmethod
    def _fit_single_grain_impl(self,
                               series_values: time_series_data_frame.TimeSeriesDataFrame,
                               grain_level: GrainType) -> Any:
        """
        Return a fitted model for a single timeseries.

        :param series_values:
            an array that represents the timeseries.
        :param grain_level:
            tuple that identifies the timeseries the model belongs to.
        :Returns:
            a model object that can be used to make predictions.
        """
        raise NotImplementedError


class AutoArima(_MultiGrainForecastBase):
    """AutoArima multigrain forecasting model."""

    def __init__(self, **kwargs):
        """Create an autoarima multi-grain forecasting model."""
        import pmdarima
        self.pmdarima = pmdarima

        timeseries_param_dict = kwargs[constants.TimeSeriesInternal.TIMESERIES_PARAM_DICT]
        super().__init__(timeseries_param_dict)

    def __getstate__(self):
        """Get state for pickle. Removes the pmdarima module object."""
        state = self.__dict__.copy()
        del state['pmdarima']
        return state

    def __setstate__(self, state):
        """Set state for pickle. Adds back the pmdarima module object."""
        self.__dict__.update(state)
        import pmdarima
        self.pmdarima = pmdarima

    def _fit_in_sample_single_grain_impl(self,
                                         model: Any,
                                         grain_level: Tuple[Any],
                                         X_grain: time_series_data_frame.TimeSeriesDataFrame) -> np.ndarray:
        date_filter = X_grain.time_index.values
        date_argmax = date_filter.argmax()
        date_range = pd.date_range(
            start=self._first_observation_dates[grain_level],
            end=date_filter[date_argmax],
            freq=self._freq)

        index = np.searchsorted(date_range, date_filter)
        if model.seasonal_order is None:
            # When seasonal_order isn't set in the pmdarima model, we should be dealing with statsmodel Arima
            # instead of SARIMAX
            # statsmodels is buggy - so catch exceptions here and default to returning zeros for in-sample preds
            # in-sample predictions are not essential for selection or forecasting so this is the least bad option
            # Don't return NaNs because the runner fails if predictions contain NaN values.
            try:
                n_ar, n_diff, _ = model.order
                if n_diff > 0:
                    # pmdarima can only return predictions from the differenced series here, so call statsmodels
                    # ARIMAResults object directly with the 'levels' arg to force the desired output
                    pred = model.arima_res_.predict(typ='levels')
                else:
                    pred = model.arima_res_.predict()
                n_padding = n_ar + n_diff
                if n_padding > 0:
                    # ARIMA predictions aren't available for the beginning of the series if the model
                    # has autoregressive components and/or has been differenced, so pad the beginning with zeros
                    # in order to align prediction output with the date grid
                    padding = np.zeros(n_padding)
                    pred = np.concatenate((padding, pred))
            except Exception:
                pred = np.zeros(date_range.shape[0])
        else:
            pred = model.predict_in_sample(start=0, end=date_range.shape[0])

        # In case of unforeseen statsmodels bugs around in-sample prediction,
        # check if we will request invalid indices and prepend zeros if so.
        max_index = index.max()
        if pred.size <= max_index:
            n_more_padding = max_index - pred.size + 1
            more_padding = np.zeros(n_more_padding)
            pred = np.concatenate((more_padding, pred))

        return cast(np.ndarray, pred[index])

    def _get_forecast_single_grain_impl(self,
                                        model: Any,
                                        max_horizon: int,
                                        grain_level: Tuple[Any],
                                        X_pred_grain: time_series_data_frame.TimeSeriesDataFrame) -> np.ndarray:

        # ARIMA (unlike Prophet) needs to have a meaningful max horizon
        if max_horizon <= 0:
            raise exceptions.DataException(model_wrappers.ForecastingPipelineWrapper.FATAL_NONPOSITIVE_HORIZON,
                                           has_pii=False)

        if len(X_pred_grain.columns) > 1:
            import warnings
            warnings.warn('ARIMA(not-X) ignoring extra features, only predicting from the target')

        pred = model.predict(n_periods=int(max_horizon))

        date_filter = X_pred_grain.time_index.values
        date_min = date_filter.min()
        date_range = pd.date_range(start=date_min, periods=max_horizon, freq=self._freq)
        index = np.searchsorted(date_range, date_filter)

        return cast(np.ndarray, pred[index])

    def _fit_single_grain_impl(self, X_pred_grain: pd.DataFrame, grain_level: GrainType) -> Any:

        # Let's warn for now, eventually we'll get the metadata on what the
        # target column is (if different from dummy) and then we can decide
        #  to ignore the rest or incorporate into ARIMAX
        if len(X_pred_grain.columns) > 1:
            import warnings
            warnings.warn('ARIMA can only predict from training data forward and does not take extra features')

        series_values = X_pred_grain[constants.TimeSeriesInternal.DUMMY_TARGET_COLUMN].values.astype(float)
        if len(series_values) < constants.TimeSeriesInternal.ARIMA_TRIGGER_CSS_TRAINING_LENGTH:
            # for short series, uses MLE training via statsmodels SARIMAX class
            model = self.pmdarima.auto_arima(series_values, error_action="ignore")
        else:
            # for long series, use CSS training (faster for longer series, less accurrate)
            # note: as of pmdarima 1.5.1, SARIMAX only is provided (no CSS option, slower training)
            model = self.pmdarima.auto_arima(series_values, error_action="ignore", seasonal=False, method="CSS")
        return model


class ProphetModel(_MultiGrainForecastBase):
    """Prophet multigrain forecasting model."""
    # "ds" and "y" are hard coded name of time and target columns in prophet package.
    PROPHET_TIME_COLUMN = 'ds'
    PROPHET_TARGET_COLUMN = 'y'

    def __init__(self, **kwargs: Any):
        """
        Construct an instance of the Prophet forecasting model.

        Accepts external regressors, which must be numeric-encoded
        (using LabelEncode, OneHotEncode, whatever, but not actual categoricals).

        Prophet parameters are accepted as a dictionary passed as a kwarg with key
        'prophet_param_dict' (or what `constants.TimeSeriesInternal.PROPHET_PARAM_DICT` says)
        and value being the actual dictionary of Prophet parameters (?help Prophet)
        """
        self.fbprophet = import_utilities.import_fbprophet()

        timeseries_param_dict = kwargs[constants.TimeSeriesInternal.TIMESERIES_PARAM_DICT]
        super().__init__(timeseries_param_dict)

        self.prophet_param_dict = {}    # type: Dict[str, Union[int, str]]
        if constants.TimeSeriesInternal.PROPHET_PARAM_DICT in kwargs.keys():
            self.prophet_param_dict = kwargs[constants.TimeSeriesInternal.PROPHET_PARAM_DICT]

        # set uncertainty_samples=0 to speed up prophet forecasting step,
        # the point forecast won't be affected and would remain exactly the same
        self.prophet_param_dict['uncertainty_samples'] = 0

        # for saving the prediction intervals
        self._pred_intervals = dict()   # type: Dict[float, Any]

    def __getstate__(self):
        """Get state for pickle. Removes the fbprophet module object."""
        state = self.__dict__.copy()
        del state['fbprophet']
        return state

    def __setstate__(self, state):
        """Set state for pickle. Adds back the fbprophet module object."""
        self.__dict__.update(state)
        self.fbprophet = import_utilities.import_fbprophet()

    def _fit_in_sample_single_grain_impl(self,
                                         model: Any,
                                         grain_level: Tuple[Any],
                                         X_grain: time_series_data_frame.TimeSeriesDataFrame) -> np.ndarray:
        fitted = self._get_forecast_single_grain_impl(model, 0, grain_level, X_grain)
        return fitted

    def _rename_time_column(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Rename the time column so that it will correspond to prophet notation ("ds").

        *Note:* This method does not create a copy of a data frame.
        :param X: The data frame.
        :return: The data frame with time column replaced to "ds".
        """
        if isinstance(X.index, pd.MultiIndex):
            # Data frame has time and grain columns.
            X.index.rename(names=ProphetModel.PROPHET_TIME_COLUMN,
                           level=self.time_column_name, inplace=True)
        elif isinstance(X.index, pd.DatetimeIndex):
            # Data frame has only time column.
            X.index.rename(ProphetModel.PROPHET_TIME_COLUMN, inplace=True)
        else:
            # Range index. Just rename the time column.
            X.rename(columns={self.time_column_name: ProphetModel.PROPHET_TIME_COLUMN},
                     inplace=True)
        return X

    def _get_forecast_single_grain_impl(self,
                                        model: Any,
                                        max_horizon: int,
                                        grain_level: Tuple[Any],
                                        X_pred_grain: time_series_data_frame.TimeSeriesDataFrame) -> np.ndarray:
        """
        Forecast from Prophet model.

        Respects the X_pred_grain parameter, rather than the max_horizon parameter.
        """
        df = pd.DataFrame(X_pred_grain.copy())
        df = self._rename_time_column(df)
        if constants.TimeSeriesInternal.DUMMY_GRAIN_COLUMN in df.columns and \
                constants.TimeSeriesInternal.DUMMY_GRAIN_COLUMN in df.index.names:
            df.drop(constants.TimeSeriesInternal.DUMMY_GRAIN_COLUMN, axis=1, inplace=True)

        df.reset_index(inplace=True)

        # grain = tuple(df.iloc[0][self.grain_column_names])
        # print("INFO: Scoring Prophet on " + str(grain))

        # TODO: how do forecast with Prophet if the user provides recent values
        # of y in a recent context?

        preds = model.predict(df)

        # TODO: save the effects from the output.  The effect is defined as:
        # "yhat is <effect> higher due to the actual value of the extra
        # regressor
        # than it would be with the "baseline" value of the regressor
        # (which would be the mean if it was standardized as by default).

        # for predict_quantiles, predictive_samples() should be used instead of
        # predict,
        # and quantiles computed on that output

        if len(self._quantiles) == 1 and self._quantiles[0] == 0.5:
            if not X_pred_grain.time_index.is_monotonic:
                # if the input was not sorted, prophet will sort it for us.
                # we need to un-do the reordering in this case.
                index = X_pred_grain.time_index.values.argsort().argsort()
                return cast(np.ndarray, preds['yhat'].values[index])
            else:
                return cast(np.ndarray, preds['yhat'].values)
        else:
            raise NotImplementedError("Quantile forecasts from Prophet are not yet supported")

    def _fit_single_grain_impl(self,
                               series_frame: pd.DataFrame,
                               grain_level: GrainType) -> Any:
        """Fit prophet model on one grain."""
        # get the data in the shape prophet expects: 'ds' for time and 'y' for
        # target

        # also, series_frame is a TSDF pandas dataframe with time and grain in
        # multiindex
        df = pd.DataFrame(series_frame.copy())
        df = self._rename_time_column(df)

        if constants.TimeSeriesInternal.DUMMY_GRAIN_COLUMN in df.columns and \
                constants.TimeSeriesInternal.DUMMY_GRAIN_COLUMN in df.index.names:
            df.drop(constants.TimeSeriesInternal.DUMMY_GRAIN_COLUMN, axis=1, inplace=True)

        df.reset_index(inplace=True)

        df.rename(
            columns={constants.TimeSeriesInternal.DUMMY_TARGET_COLUMN: ProphetModel.PROPHET_TARGET_COLUMN},
            inplace=True)

        cols_to_ignore = [ProphetModel.PROPHET_TIME_COLUMN,
                          ProphetModel.PROPHET_TARGET_COLUMN] + self.grain_column_names + self.drop_column_names
        list_of_features = [x for x in df.columns if x not in cols_to_ignore]
        # grain = tuple(df.iloc[0][self.grain_column_names])
        # print("INFO: Fitting Prophet on " + str(grain)) # doesn't print anyway even with pytest -s

        model = self.fbprophet.Prophet(**self.prophet_param_dict)
        if len(list_of_features) > 0:
            for feat in list_of_features:
                model.add_regressor(feat)

        # Catch the FutureWarning from Prophet
        with utilities.suppress_stdout_stderr():
            model.fit(df)

        return model
